"""THIS SOFTWARE IS PROVIDED AS IS.

Released under GNU General Public License:
<https://www.gnu.org/licenses/gpl-3.0.en.html>

USE IT AT YOUR OWN RISK.

PipHyperd unit testing.

The module is published on PyPi: <https://pypi.org/project/piphyperd/>.
The code is available on GitLab: <https://gitlab.com/hyperd/piphyperd>.
"""

import unittest
import os
import shutil
from pathlib import Path
import subprocess
from ..main.piphyperd import PipHyperd


class TestMethods(unittest.TestCase):
    """PipHyperd unit testing class."""

    def setUp(self) -> None:
        """Test setup."""
        self.piphyperd = PipHyperd()
        self.venv_path = "{}/python-venv".format(os.path.dirname(__file__))

        subprocess.run(
            ["virtualenv",
             self.venv_path], check=True,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    def tearDown(self) -> None:
        """Remove venv after testing."""
        self.wiper(self.venv_path)

    def test_a_is_not_none(self) -> None:
        """Assert that PipHyperd is not None."""
        self.assertIsNotNone(self.piphyperd)

    def test_b_wrong_python_path(self) -> None:
        """Raise a FileNotFoundError when python path does not exist."""
        with self.assertRaises(FileNotFoundError):
            PipHyperd(python_path=Path("/path/to/nothing")).check()

    def test_c_install(self) -> None:
        """Assert that after installing is in the output."""
        self.piphyperd = PipHyperd(
            python_path=Path("{}/bin/python3".format(self.venv_path)))

        self.piphyperd.install("ansible")

        output, _, _ = self.piphyperd.list_packages()

        self.assertIn("ansible", output)

    def test_d_uninstall(self) -> None:
        """Assert that after installing is in the output."""
        self.piphyperd = PipHyperd(
            python_path=Path("{}/bin/python3".format(self.venv_path)))

        self.piphyperd.uninstall("ansible")

        output, _, _ = self.piphyperd.list_packages()

        self.assertNotIn("ansible", output)

    def test_e_list_outdated(self) -> None:
        """Assert that "Latest" is in the output."""
        self.piphyperd = PipHyperd(
            python_path=Path("{}/bin/python3".format(self.venv_path)))

        # install an outdated version of yarl
        self.piphyperd.install("yarl==1.1.0")

        output, _, _ = self.piphyperd.list_packages(True)

        self.assertIn("Latest", output)

    @staticmethod
    def wiper(folder: str) -> None:
        """Clean up folders."""
        if os.path.isdir(folder):
            shutil.rmtree(folder)


if __name__ == '__main__':
    unittest.main()



"""THIS SOFTWARE IS PROVIDED AS IS.

Released under GNU General Public License:
<https://www.gnu.org/licenses/gpl-3.0.en.html>

USE IT AT YOUR OWN RISK.

Command proxy for the PipHyperd cli.
piphyperd is a wrapper around **pip**; it can provide features like
automation or dependencies control within your workflows.

The module is published on PyPi: <https://pypi.org/project/piphyperd/>.
The code is available on GitLab: <https://gitlab.com/hyperd/piphyperd>.
"""

from typing import Tuple, Any
from ..main.piphyperd import PipHyperd


class CmdProxy:
    """PipHyperd interface class."""

    @staticmethod
    def freeze(instance: PipHyperd) -> Tuple[str, str, int]:
        """Return pip freeze."""
        return instance.freeze()

    @staticmethod
    def list_packages(instance: PipHyperd) -> Tuple[str, str, int]:
        """Return pip list."""
        return instance.list_packages()

    @staticmethod
    def show(instance: PipHyperd, packages: Any) -> Tuple[str, str, int]:
        """Return pip show ${package}."""
        return instance.show(packages[0])

    @staticmethod
    def check(instance: PipHyperd) -> Tuple[str, str, int]:
        """Return pip check."""
        return instance.check()

    @staticmethod
    def dependencies_tree(instance: PipHyperd) -> Tuple[str, str, int]:
        """Return pip check."""
        return instance.dependencies_tree()

    @staticmethod
    def install(instance: PipHyperd, packages: Any) -> Tuple[str, str, int]:
        """Return pip install ${packages}."""
        return instance.install(*packages)

    @staticmethod
    def uninstall(instance: PipHyperd, packages: Any) -> Tuple[str, str, int]:
        """Return pip uninstall ${packages}."""
        return instance.uninstall(*packages)

    @staticmethod
    def download(instance: PipHyperd, packages: Any) -> Tuple[str, str, int]:
        """Return pip download ${packages}."""
        return instance.download(*packages)



"""THIS SOFTWARE IS PROVIDED AS IS.

Released under GNU General Public License:
<https://www.gnu.org/licenses/gpl-3.0.en.html>

USE IT AT YOUR OWN RISK.

PipHyperd is a simple python object to leverage pip programmatically.
It can provide automation and dependencies control within your workflows.

The module is published on PyPi: <https://pypi.org/project/piphyperd/>.
The code is available on GitLab: <https://gitlab.com/hyperd/piphyperd>.
"""

import subprocess
import sys
from pathlib import Path
from typing import Optional, List, Tuple, Any, Union


class PipHyperd:
    """Wrapper class around pip.

    python_path -- Path to the python binary to use

    *pip_options -- command options, e.g.: pip {pip_options} uninstall testpypi
    """

    def __init__(self, *pip_options: Any, python_path: Optional[Path] = None):
        """Init method."""
        # Path to the python binary to use
        self.python_path: Optional[Path] = python_path
        # A list of pip packages to install || show || download || uninstall
        self.packages: List[str] = list()
        # pip command args, e.g.: pip download testpypi {command_args}
        self.command_args: List[str] = list()
        # pip options, e.g.: pip {pip_options} uninstall testpypi
        self.pip_options: List[str] = list(pip_options)

    def __subprocess_wrapper(self, command: str,
                             wait: bool = False) -> Tuple[str, str, int]:
        """Subprocess wrapper allowing to execute pip commands.

        command -- The pip command to execute (e.g. "install", "freeze", [..])
        wait -- True || False wait for the process to terminate
        """
        try:
            # leverage subprocess.Popen to execute pip commands
            pip_full_cmd: Union[Any] = sorted(
                self.pip_options + self.packages + self.command_args)

            process = subprocess.Popen(
                [sys.executable if self.python_path is None
                 else self.python_path,
                 "-m", "pip", command] + pip_full_cmd,
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)

            # wait for the process to terminate
            if wait:
                process.wait()

            stdout, stderr = process.communicate()

            output = f'{stdout.decode("utf-8")}'
            sys.stdout.write(output)

            outerr = f'{stderr.decode("utf-8")}'
            sys.stderr.write(outerr)

            return output, outerr, process.returncode

        except subprocess.CalledProcessError as called_process_error:
            print(called_process_error)
            ex_output: str = f'Error output:\n{called_process_error.output}'
            ex_cmd: str = f'cmd:\n{called_process_error.cmd}'
            return ex_output, ex_cmd, called_process_error.returncode

    def freeze(self) -> Tuple[str, str, int]:
        """Return installed pip packages in requirements format."""
        return self.__subprocess_wrapper("freeze", wait=True)

    def list_packages(self,
                      list_outdated: bool = False) -> Tuple[str, str, int]:
        """List installed pip packages.

        list_outdated -- True || False to list or not the outdated packages
        """
        if list_outdated:
            self.command_args.insert(0, "--outdated")

        return self.__subprocess_wrapper("list")

    def show(self, package: str) -> Tuple[str, str, int]:
        """Show information about installed packages.

        package -- Package to show
        """
        self.packages.append(package)
        return self.__subprocess_wrapper("show")

    @staticmethod
    def dependencies_tree() -> Tuple[str, str, int]:
        """List a per-package dependencies tree."""
        process = subprocess.run(
            [str(sys.executable),
             "-m", "pipdeptree", "--json-tree"], check=True,
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        process.check_returncode()

        stdout = f'{process.stdout.decode("utf-8")}'
        sys.stdout.write(stdout)

        stderr = f'{process.stderr.decode("utf-8")}'
        sys.stderr.write(stderr)

        return stdout, stderr, process.returncode

    def check(self) -> Tuple[str, str, int]:
        """Verify installed packages have compatible dependencies."""
        return self.__subprocess_wrapper("check", wait=True)

    def install(self, *packages: Any) -> Tuple[str, str, int]:
        """Install pip packages.

        *packages -- List of packages to install
        """
        self.packages.clear()

        for package in packages:
            self.packages.append(str(package))

        return self.__subprocess_wrapper("install")

    def uninstall(self, *packages: Any) -> Tuple[str, str, int]:
        """Uninstall pip packages.

        *packages -- List of packages to uninstall
        """
        self.packages.clear()

        for package in packages:
            self.packages.append(str(package))

        # self.pip_options.insert(0, "-y")
        return self.__subprocess_wrapper("uninstall")

    def download(self, *packages: Any,
                 destination: Optional[Path] = None) -> Tuple[str, str, int]:
        """Download pip packages.

        *packages -- List of packages to download

        destination -- Destination path for the packages download
        """
        self.packages.clear()

        for package in packages:
            self.packages.append(str(package))

        if destination is not None:
            destination = Path(str(destination).strip())
            self.command_args.append("-d{}".format(destination))

        return self.__subprocess_wrapper("download")



"""THIS SOFTWARE IS PROVIDED AS IS.

Released under GNU General Public License:
<https://www.gnu.org/licenses/gpl-3.0.en.html>

USE IT AT YOUR OWN RISK.

PipHyperd is a simple python object to leverage pip programmatically.
It can provide automation and dependencies control within your workflows.

The module is published on PyPi: <https://pypi.org/project/piphyperd/>.
The code is available on GitLab: <https://gitlab.com/hyperd/piphyperd>.
"""
from __future__ import (absolute_import, division, print_function)

__metaclass__ = type  # pylint: disable=invalid-name

__version__ = "1.9.6"
__author__ = "Hyper(d)"



"""THIS SOFTWARE IS PROVIDED AS IS.

Released under GNU General Public License:
<https://www.gnu.org/licenses/gpl-3.0.en.html>

USE IT AT YOUR OWN RISK.

PipHyperd setup.
It can provide automation and dependencies control within your workflows.

The module is published on PyPi: <https://pypi.org/project/piphyperd/>.
The code is available on GitLab: <https://gitlab.com/hyperd/piphyperd>.
"""

from __future__ import (absolute_import, division, print_function)
import codecs
import os
import re
from typing import Any
import setuptools

entry_points: str = """\"""THIS SOFTWARE IS PROVIDED AS IS.

Released under GNU General Public License:
<https://www.gnu.org/licenses/gpl-3.0.en.html>

USE IT AT YOUR OWN RISK.

PipHyperd is a simple python object to leverage pip programmatically.
It can provide automation and dependencies control within your workflows.

The module is published on PyPi: <https://pypi.org/project/piphyperd/>.
The code is available on GitLab: <https://gitlab.com/hyperd/piphyperd>.
\"""
from __future__ import (absolute_import, division, print_function)

__metaclass__ = type  # pylint: disable=invalid-name

__version__ = \"{version}\"
__author__ = \"Hyper(d)\"
"""


def envstring(var: str) -> str:
    """Return environment var as string."""
    return os.environ.get(var) or ""


def read(rel_path: str) -> Any:
    """Read file helper."""
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(here, rel_path), 'r') as file_path:
        return file_path.read()


try:
    LONG_DESCRIPTION = read("README.md")
except FileNotFoundError:
    LONG_DESCRIPTION = ""

if os.path.isfile("variables"):
    try:
        VARIABLES = read("variables").strip().split("\n")
        for v in VARIABLES:
            key, value = v.split("=")
            os.environ[key] = re.sub("['\"]", "", value)
    except FileNotFoundError:
        pass


setuptools.setup(
    name=envstring("NAME"),

    # Version SCM based.
    use_scm_version={
        'write_to': 'module/main/release.py',
        'write_to_template': entry_points,
        'tag_regex': r'^(?P<prefix>v)?(?P<version>[^\+]+)(?P<suffix>.*)?$',
        'version_scheme': 'python-simplified-semver',
        'local_scheme': 'no-local-version',
    },

    description=envstring("DESCRIPTION"),
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",

    license='GPLv3',

    author=envstring("AUTHOR"),
    author_email=envstring("AUTHOR_EMAIL"),

    url=envstring("URL"),

    project_urls={
        "Documentation": envstring("URL"),
        "Source": envstring("DOCUMENTATION_URL"),
    },

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
        'Environment :: Plugins',
    ],

    install_requires=[
        'pipdeptree==0.13.2',
    ],

    packages=[
        envstring("NAME"), envstring("NAME") + ".main",
        envstring("NAME"), envstring("NAME") + ".cli",
    ],

    entry_points={
        "console_scripts": [
            "piphyperd=piphyperd.cli.parser:run",
        ]
    },

    python_requires='>=3.5.*',
    zip_safe=False,
)



"""THIS SOFTWARE IS PROVIDED AS IS.

Released under GNU General Public License:
<https://www.gnu.org/licenses/gpl-3.0.en.html>

USE IT AT YOUR OWN RISK.

PipHyperd cli interface.

The module is published on PyPi: <https://pypi.org/project/piphyperd/>.
The code is available on GitLab: <https://gitlab.com/hyperd/piphyperd>.
"""

import sys
import argparse
from typing import Optional, List, Any
from pathlib import Path
from ..main.piphyperd import PipHyperd
from .cmdproxy import CmdProxy


def run(python_path: Optional[Path] = None, command: str = "") -> int:
    """Execute when run as python3 -m ${MODULE}.

    Parse any additional arguments and call required module functions.
    """
    if sys.argv:
        # called through CLI
        # module_name = __loader__.name.split('.')[0]
        parser = argparse.ArgumentParser(
            prog="piphyperd",
            description="A pip wrapper to manage packages in your workflows.",
        )

        parser.add_argument('--python_path', action='store', nargs=1,
                            required=False, type=str,
                            default=[sys.executable],
                            help="Path to a python binary'")

        parser.add_argument('command', nargs=1, type=str,
                            default=["list"],
                            help="Provide a valid pip command")

        parser.add_argument('--packages', action="store", nargs="+",
                            required=False, type=str,
                            help="Provide a list of packages")

        # argparser provides us a list, even if only one argument
        args = parser.parse_args(sys.argv[1:])

        # check for alternative python path used to initialize the
        if args.python_path:
            if isinstance(args.python_path,
                          list) and isinstance(args.python_path[0], str):
                python_path = Path(args.python_path[0])

        # check for alternative python path
        if args.command:
            if isinstance(args.command,
                          list) and isinstance(args.command[0], str):
                command = args.command[0]

        command_args: List[str] = []
        if args.packages:
            command_args = list()
            if isinstance(args.packages,
                          list) and isinstance(args.packages[0], str):
                command_args.clear()
                for package in args.packages:
                    command_args.append(package)

        switcher = {
            "freeze": CmdProxy.freeze,
            "list": CmdProxy.list_packages,
            "show": CmdProxy.show,
            "check": CmdProxy.check,
            "install": CmdProxy.install,
            "uninstall": CmdProxy.uninstall,
            "download": CmdProxy.download,
            "dependencies-tree": CmdProxy.dependencies_tree,
        }

        func: Any = switcher.get(command, lambda: 'Invalid command')

    instance = PipHyperd(python_path=python_path)

    _, _, exit_code = func(
        instance) if not command_args else func(instance, command_args)

    return int(exit_code)
